### Script written during ChessAgent + PythonScriptingAgent Insightful experiment

#!/usr/bin/env python3

import argparse
import chess
import chess.engine
import json
import sys
import time
import asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, List, Optional, Tuple
from contextlib import contextmanager

# Constants for analysis
ANALYSIS_DEPTHS = [20, 25, 30]
MAX_ANALYSIS_TIME = 0.1  # seconds per analysis to prevent timeouts
CONFIDENCE_THRESHOLD = 0.8
MAX_ALTERNATIVE_LINES = 3
ENGINE_TIMEOUT = 30  # maximum time to wait for engine operations

def create_engine():
    """Create and configure the chess engine with proper error handling."""
    try:
        engine = chess.engine.SimpleEngine.popen_uci("stockfish")
        engine.configure({"Threads": 2, "Hash": 128})
        return engine
    except (chess.engine.EngineTerminatedError, FileNotFoundError) as e:
        print(f"Error: Failed to initialize chess engine - {str(e)}", file=sys.stderr)
        sys.exit(1)

def safe_quit_engine(engine):
    """Safely quit the engine with proper error handling."""
    if engine:
        try:
            engine.quit()
        except Exception as e:
            print(f"Warning: Error while closing engine - {str(e)}", file=sys.stderr)

def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description='Validate mate sequences with multi-depth analysis')
    parser.add_argument('--fen', type=str, required=True,
                       help='FEN string of the position to analyze')
    parser.add_argument('--candidate_move', type=str, required=True,
                       help='First move of potential mating sequence in UCI notation')
    parser.add_argument('--num_variations', type=int, required=True,
                       help='Number of top defensive tries to analyze (1-5)')
    
    args = parser.parse_args()
    
    if not (1 <= args.num_variations <= 5):
        print("Error: num_variations must be between 1 and 5", file=sys.stderr)
        sys.exit(1)
        
    return args

def analyze_position_at_depth(engine: chess.engine.SimpleEngine,
                            board: chess.Board,
                            depth: int,
                            multipv: int) -> List[Dict]:
    """Analyze a position at a specific depth with proper error handling."""
    try:
        # Set a strict time limit for analysis and include multipv in Limit
        limit = chess.engine.Limit(depth=depth, time=MAX_ANALYSIS_TIME)
        info = engine.analyse(
            board,
            limit=limit,
            multipv=multipv
        )
        
        return [{
            "depth": pv.get("depth", depth),
            "score": str(pv["score"].relative),
            "mate": pv["score"].relative.mate(),
            "pv": [move.uci() for move in pv.get("pv", [])]
        } for pv in info]
    except (chess.engine.EngineError, chess.engine.EngineTerminatedError) as e:
        print(f"Warning: Engine error at depth {depth}: {str(e)}", file=sys.stderr)
        return []

def validate_mate_sequence(board: chess.Board,
                         engine: chess.engine.SimpleEngine,
                         candidate_move: chess.Move,
                         num_variations: int) -> Dict:
    """Validate a potential mate sequence through multi-depth analysis."""
    results = {depth: [] for depth in ANALYSIS_DEPTHS}
    confidence_scores = []
    
    # Make the candidate move
    if candidate_move not in board.legal_moves:
        return {
            "valid_mate": False,
            "error": "Candidate move is not legal"
        }
    
    board.push(candidate_move)
    
    # Analyze position after candidate move at different depths
    with ThreadPoolExecutor(max_workers=len(ANALYSIS_DEPTHS)) as executor:
        future_to_depth = {
            executor.submit(
                analyze_position_at_depth,
                engine,
                board.copy(),
                depth,
                num_variations
            ): depth for depth in ANALYSIS_DEPTHS
        }
        
        for future in as_completed(future_to_depth):
            depth = future_to_depth[future]
            try:
                analysis = future.result()
                if analysis:  # Only store non-empty results
                    results[depth] = analysis
            except Exception as e:
                print(f"Error analyzing at depth {depth}: {str(e)}", file=sys.stderr)
    
    # Process results and calculate confidence
    mate_lengths = []
    defensive_resources = []
    
    for depth in ANALYSIS_DEPTHS:
        if results[depth]:
            main_line = results[depth][0]
            
            # Check for mate
            if main_line["mate"] is not None:
                mate_lengths.append(abs(main_line["mate"]))
                confidence_scores.append(1.0)
            else:
                confidence_scores.append(0.0)
            
            # Collect defensive resources
            for variation in results[depth][1:]:
                if variation not in defensive_resources:
                    defensive_resources.append(variation)
    
    # Calculate overall confidence with safety check
    overall_confidence = 0.0
    if confidence_scores:
        overall_confidence = sum(confidence_scores) / len(confidence_scores)
    
    # Determine if it's a valid mate sequence
    is_valid_mate = overall_confidence >= CONFIDENCE_THRESHOLD
    mate_length = min(mate_lengths) if mate_lengths else None
    
    return {
        "valid_mate": is_valid_mate,
        "mate_length": mate_length,
        "confidence_score": round(overall_confidence, 3),
        "analysis_by_depth": {
            str(depth): results[depth]
            for depth in ANALYSIS_DEPTHS
            if results[depth]  # Only include depths with valid results
        },
        "defensive_resources": defensive_resources[:num_variations],
        "alternative_lines": results[max(results.keys())][:MAX_ALTERNATIVE_LINES] if results else []
    }

def main():
    """Main function to validate mate sequences with proper error handling."""
    args = parse_arguments()
    engine = None
    
    try:
        board = chess.Board(args.fen)
    except ValueError as e:
        print(f"Error: Invalid FEN string - {str(e)}", file=sys.stderr)
        sys.exit(1)
    
    try:
        candidate_move = chess.Move.from_uci(args.candidate_move)
    except ValueError as e:
        print(f"Error: Invalid UCI move - {str(e)}", file=sys.stderr)
        sys.exit(1)
    
    try:
        engine = create_engine()
        analysis = validate_mate_sequence(
            board,
            engine,
            candidate_move,
            args.num_variations
        )
        print(json.dumps(analysis, indent=2))
    except Exception as e:
        print(f"Error during analysis: {str(e)}", file=sys.stderr)
        sys.exit(1)
    finally:
        if engine:
            safe_quit_engine(engine)

if __name__ == "__main__":
    main()